; Testen, ob eine Atomformel durch ein "builtin" Praedikat behandelt wird.
(define builtin?
  (lambda (atomformel)
    (or (and (bezeichner-passt? atomformel "plus")
	     (stelligkeit-gleich? atomformel 3))
	(and (bezeichner-passt? atomformel "mal")
	     (stelligkeit-gleich? atomformel 3))
	(and (bezeichner-passt? atomformel "positiv")
	     (stelligkeit-gleich? atomformel 1)))))

; Hat eine Atomformel eine bestimmte Stelligkeit?
(define stelligkeit-gleich?
  (lambda (atomformel stelligkeit)
    (= (length (atomformel->terme atomformel)) stelligkeit)))

; Stimmt das Praed-Symbol einer Atomformel mit einem best. String ueberein?
(define bezeichner-passt?
  (lambda (atomformel string)
    (praedikatensymbol=?
     (make-praedikatensymbol string)
     (atomformel->praedikatensymbol atomformel))))

; Ist die Atomformel war in der Datenbasis? Antwort fuer builtin-Praedikate.
(define builtin-atomformel-in-datenbasis?
  (lambda (atomformel)
    (cond ((bezeichner-passt? atomformel "plus")
	   (plus? (atomformel->terme atomformel)))
	  ((bezeichner-passt? atomformel "mal")
	   (mal? (atomformel->terme atomformel)))
	  ((bezeichner-passt? atomformel "positiv")
	   (positiv? (car (atomformel->terme atomformel))))
	  (else #f))))

; Finde alle (im allgemeinen hier: die einzige oder keine) Substitutionen
;  fuer eine Atomformel mit Variablen
(define builtin-passende-substitutionen
  (lambda (atomformel)
    (cond ((bezeichner-passt? atomformel "plus")
	   (plus-substitutionen (atomformel->terme atomformel)))
	  ((bezeichner-passt? atomformel "mal")
	   (mal-substitutionen (atomformel->terme atomformel)))
	  ((bezeichner-passt? atomformel "positiv")
	   (positiv-substitutionen (atomformel->terme atomformel)))
	  (else '()))))

; Test, ob plus/3 Fakt erfuellt ist.
(define plus?
  (lambda (terme) (if (null? (plus-muster-zzz terme)) #f #t)))

; Test, ob mal/3 Fakt erfuellt ist.
(define mal?  
  (lambda (terme) (if (null? (mal-muster-zzz terme))  #f #t)))

; Finde alle Substitutionen fuer plus(_,_,_): Eine oder keine S.
(define plus-substitutionen
  (lambda (terme)
    (let ((im (instanziierungsmuster terme)))
      (cond ((string=? im "VVV") (plus-muster-vvv terme))
	    ((string=? im "VVZ") (plus-muster-vvz terme))
	    ((string=? im "VZV") (plus-muster-vzv terme))
	    ((string=? im "ZVV") (plus-muster-zvv terme))
	    ((string=? im "ZZV") (plus-muster-zzv terme))
	    ((string=? im "ZVZ") (plus-muster-zvz terme))
	    ((string=? im "VZZ") (plus-muster-vzz terme))
	    ((string=? im "ZZZ") (plus-muster-zzz terme))
	    (else                '()                    )))))

; Finde alle Substitutionen fuer mal(_,_,_): Eine oder keine S.
(define mal-substitutionen
  (lambda (terme)
    (let ((im (instanziierungsmuster terme)))
      (cond ((string=? im "VVV") (mal-muster-vvv terme))
	    ((string=? im "VVZ") (mal-muster-vvz terme))
	    ((string=? im "VZV") (mal-muster-vzv terme))
	    ((string=? im "ZVV") (mal-muster-zvv terme))
	    ((string=? im "ZZV") (mal-muster-zzv terme))
	    ((string=? im "ZVZ") (mal-muster-zvz terme))
	    ((string=? im "VZZ") (mal-muster-vzz terme))
	    ((string=? im "ZZZ") (mal-muster-zzz terme))
	    (else                '()                    )))))

; Finde alle Substitutionen fuer positiv/1. Maximal eine: die leere S.
(define positiv-substitutionen
  (lambda (terme)
    (if (variable? (car terme))
	(error (apply string-append (cons "Instanziierungsfehler positiv("
					  (unparse-term (car terme) '(")")))))
	(if (positiv? (car terme))
	    (list leere-substitution)
	    '()))))


; terme muss eine Liste von drei Variablen sein. Sind sie alle gleich?
(define drei-variablen-gleich?
  (lambda (terme)
    (and (variable=? (car terme) (cadr terme))
	  (variable=? (car terme) (caddr terme)))))

; Jetzt kommen die Funktionen, die fuer plus/3 und mal/3 die Substitutionen
; berechnen. Fuer jedes der acht moeglichen Instanziierungsmuster gibt es
; eine Funktion fuer mal und fuer plus. Insgesamt also 16 Funktionen.
; VVV -----------------------------------------------------------------------
(define plus-muster-vvv
  (lambda (terme)
    (if (drei-variablen-gleich? terme)
	(list (substitution-erweitern
	       (car terme)
	       (zahl->term 0)
	       leere-substitution))
	(display-nl "Ungltiges Instanziierungsmuster plus(X,Y,Z)."))))

(define mal-muster-vvv
  (lambda (terme)
    (if (drei-variablen-gleich? terme)
	(list (substitution-erweitern
	       (car terme)
	       (zahl->term 1)
	       leere-substitution))
	(display-nl "Ungltiges Instanziierungsmuster mal(X,Y,Z)."))))

; ZVV ----------------------------------------------------------------------
(define plus-muster-zvv
  (lambda (terme)
    (plus-muster-vzv (list (cadr terme) (car terme) (caddr terme)))))

(define mal-muster-zvv
  (lambda (terme)
    (mal-muster-vzv (list (cadr terme) (car terme) (caddr terme)))))

; VZV ----------------------------------------------------------------------
(define plus-muster-vzv   
  (lambda (terme)
    (if (variable=? (car terme) (caddr terme))
	(if (zero? (term->zahl (cadr terme)))
	    (display-nl "Ungltiges Instanziierungsmuster plus(X,0,X) oder plus(0,X,X).")
	    '())
	(display-nl "Ungltiges Instanziierungsmuster plus(X,c,Y) oder plus(c,X,Y)."))))

(define mal-muster-vzv
  (lambda (terme)
    (if (variable=? (car terme) (caddr terme))
	(if (= 1 (term->zahl (cadr terme)))
	    (display-nl "Ungltiges Instanziierungsmuster mal(X,1,X) oder mal(1,X,X).")
	    '())
	(display-nl "Ungltiges Instanziierungsmuster mal(X,c,Y) oder mal(c,X,Y)."))))

; VVZ ------------------------------------------------------------------------
(define plus-muster-vvz
  (lambda (terme)
    (if (variable=? (car terme) (cadr terme))
	(if (geradzahlig? (caddr terme))
	    (list (substitution-erweitern
		  (car terme)
		  (haelfte-eines-terms (caddr terme))
		  leere-substitution))
	    '())
	(display-nl "Ungltiges Instanziierungsmuster plus(X,Y,c)."))))

(define mal-muster-vvz
  (lambda (terme)
    (if (variable=? (car terme) (cadr terme))
	(if (quadratzahl? (caddr terme))
	    (list (substitution-erweitern
		  (car terme)
		  (wurzel-eines-terms (caddr terme))
		  leere-substitution))
	    '())
	(display-nl "Ungltiges Instanziierungsmuster mal(X,Y,c)."))))

; ZZV ------------------------------------------------------------------------
(define plus-muster-zzv
  (lambda (terme)
    (list (substitution-erweitern 
	   (caddr terme)
	   (summe-zweier-terme (car terme) (cadr terme))
	   leere-substitution))))

(define mal-muster-zzv
  (lambda (terme)
    (list (substitution-erweitern 
	   (caddr terme)
	   (produkt-zweier-terme (car terme) (cadr terme))
	   leere-substitution))))

; ZVZ ------------------------------------------------------------------------
(define plus-muster-zvz
  (lambda (terme)
    (list (substitution-erweitern 
	   (cadr terme)
	   (differenz-zweier-terme (caddr terme) (car terme))
	   leere-substitution))))

(define mal-muster-zvz
  (lambda (terme)
    (if (teilbar? (caddr terme) (car terme))
	(list (substitution-erweitern 
	       (cadr terme)
	       (quotient-zweier-terme (caddr terme) (car terme))
	       leere-substitution))
	'())))

; VZZ ------------------------------------------------------------------------
(define plus-muster-vzz
  (lambda (terme)
    (plus-muster-zvz (list (cadr terme) (car terme) (caddr terme)))))

(define mal-muster-vzz
  (lambda (terme)
    (mal-muster-zvz (list (cadr terme) (car terme) (caddr terme)))))


; ZZZ ------------------------------------------------------------------------
(define plus-muster-zzz
  (lambda (terme)
    (if (summe-zweier-terme-ist? (car terme) (cadr terme) (caddr terme))
	(list leere-substitution)
	'())))

(define mal-muster-zzz
  (lambda (terme)
    (if (produkt-zweier-terme-ist? (car terme) (cadr terme) (caddr terme))
	(list leere-substitution)
	'())))

; Hilfsfunktionen fuer Zahlen ------------------------------------------------

; Term in Zahl umwandeln. Fehlertest gibt es keinen.
(define term->zahl
  (lambda (term)
    (string->number (funktionssymbol->bezeichner (funktion->funktionssymbol term)))))

; Macht aus einer Zahl wieder einen Term.
(define zahl->term
  (lambda (zahl)
    (make-funktion (make-funktionssymbol (number->string zahl)) '())))

; Handelt es sich bei einem Term um eine Zahl?
(define zahl?
  (lambda (term)
    (and 
     (funktion? term)
     (integer? (string->number 
		(funktionssymbol->bezeichner 
		 (funktion->funktionssymbol term)))))))

; Liefert zu einer Liste von Termen einen String, der beschreibt, an welcher
; Stelle Zahlen, Variablen oder andere Terme auftreten. Z.B. liefert die
; Funktion fuer f(12,X,X,f(g(hans)),wurst,Y) das Muster "ZVVFFV".
(define instanziierungsmuster
  (lambda (terme)
    (list->string (map (lambda (term) 
			 (cond ((variable? term)  #\V)
			       ((zahl? term) #\Z)
			       (else         #\F))) terme))))

; Tests fuer Summe und Produkt ------------------------------------------------

(define summe-zweier-terme-ist?
  (lambda (s1 s2 s) (eq? (+ (term->zahl s1) (term->zahl s2)) (term->zahl s))))

(define produkt-zweier-terme-ist?
  (lambda (f1 f2 p) (eq? (* (term->zahl f1) (term->zahl f2)) (term->zahl p))))

; Rechnungen -------------------------------------------------------------------

(define summe-zweier-terme
  (lambda (s1 s2) (zahl->term (+ (term->zahl s1) (term->zahl s2)))))

(define differenz-zweier-terme
  (lambda (s1 s2) (zahl->term (- (term->zahl s1) (term->zahl s2)))))

(define produkt-zweier-terme
  (lambda (f1 f2) (zahl->term (* (term->zahl f1) (term->zahl f2)))))

(define quotient-zweier-terme
  (lambda (z n)   (zahl->term (/ (term->zahl z) (term->zahl n)))))

(define haelfte-eines-terms
  (lambda (term) (zahl->term (/ (term->zahl term) 2))))

(define wurzel-eines-terms
  (lambda (term) (zahl->term (sqrt (term->zahl term)))))

; Tests ------------------------------------------------------------------------

(define geradzahlig?
  (lambda (term) (integer? (/ (term->zahl term) 2))))

(define quadratzahl?
  (lambda (term) (and (or (zero? (term->zahl term))
			  (positive? (term->zahl term)))
		      (exact? (sqrt (term->zahl term))))))
(define teilbar?
  (lambda (z n)
    (and (not (zero? (term->zahl n)))
	 (integer? (/ (term->zahl z) (term->zahl n))))))

(define positiv?
  (lambda (term) (positive? (term->zahl term))))
