; Definition von vier Typen zur Darstellung eines Binaeren Baumes
(define-record binbaum (knoten links rechts))
(define-record knoten (atomliste))
(define-record leerer-binbaum ())
(define-record datenbasis (datenbaum))

; Funktionen fuer plus, mal und positiv nachladen
(load "builtin.scm")

; Baum von Listen in flache Liste von Atomformeln umwandeln
(define datenbasis->liste 
  (lambda (datenbasis)
    (binbaum->liste (datenbasis->datenbaum datenbasis))))

; Datenbasis auf dem Bildschirm ausgeben
(define unparse-datenbasis
  (lambda (datenbasis accu)
    (unparse-with (string #\newline) unparse-atomformel (datenbasis->liste datenbasis) accu)))

; Konstruktor fuer die initiale Datenbasis
(define leere-datenbasis
  (make-datenbasis (make-leerer-binbaum)))

; Fuegt eine Atomformel ein. Entfernt werden kann nicht. Beim Versuch,
; plus/3, mal/3 oder positiv/1 Fakten einzufuegen wird ein Fehler erzeugt.
(define atomformel-hinzufuegen
  (lambda (atomformel datenbasis)
    (if (builtin? atomformel)
	(begin
	  (display-nl "Builtin-Praedikate werden nicht eingefuegt.")
	  'jetzt-gibts-aber-bald-einen-fehler!)
	(make-datenbasis (binbaum-einfuegen
			  atomformel
			  (datenbasis->datenbaum datenbasis))))))

; Test (aussagenlogisch), ob eine variablenfreie Atomformel enthalten ist.
(define atomformel-in-datenbasis?
  (lambda (datenbasis)
    (lambda (atomformel)
      (if (builtin? atomformel)
	  (builtin-atomformel-in-datenbasis? atomformel)
	  ((dynamisch-atomformel-in-datenbasis? datenbasis) atomformel)))))

; Dieser Teil von atomformel-in-datenbasis? ist fuer nicht-builtin.
(define dynamisch-atomformel-in-datenbasis?
  (lambda (datenbasis)
    (lambda (atomformel)
      (richtiger-knoten 
       datenbasis 
       atomformel
       (lambda (knoten) 
	 (any (lambda (af) (atomformel=? atomformel af)) 
	      (knoten->atomliste knoten)))
       #f))))

; Erzeugt alle Substitutionen, unter denen eine Atomformel in der Datenbasis
; enthalten (gueltig ist). Auch fuer builtin
(define passende-substitutionen
  (lambda (atomformel datenbasis)
    (if (builtin? atomformel)
	(builtin-passende-substitutionen atomformel)
	(dynamisch-passende-substitutionen atomformel datenbasis))))

; Dieser Teil von passende-substitutionen ist fuer nicht builtin.
; Es muss nur ein Knoten vom Baum untersucht werden: Der mit dem richtigen
; Praedikatensymbol. Die Stelligkeit wird nicht untersucht.
(define dynamisch-passende-substitutionen
  (lambda (atomformel datenbasis)
    (richtiger-knoten
     datenbasis 
     atomformel
     (lambda (knoten)
       (foldr (lambda (af substitutionen)
		(atomformel-matchen atomformel af
				    (lambda (substitution) (cons substitution substitutionen))
				    (lambda () substitutionen)))
	      '()
	      (knoten->atomliste knoten)))
     '())))

; Einfache Linearisierung des Baumes.
(define binbaum->liste
  (lambda (binbaum)
    (if (leerer-binbaum? binbaum)
	'()
	(append
	 (binbaum->liste (binbaum->links binbaum))
	 (append
	  (knoten->atomliste (binbaum->knoten binbaum))
	  (binbaum->liste (binbaum->rechts binbaum)))))))

; Fuegt eine Atomformel ein. Wenn zum Praedikatensymbol dieser Formel noch
; kein Knoten existiert, so muss ein neuer Knoten einsortiert werden. 
; Ansonsten wird einfach an die Liste des bestehenden Knotens die Formel
; angehaengt.
(define binbaum-einfuegen
  (lambda (atomformel binbaum)
    (if (leerer-binbaum? binbaum)
	(make-binbaum (make-knoten (list atomformel))
		      (make-leerer-binbaum)
		      (make-leerer-binbaum))
	(let ((symbol-neu (atomformel->praedikatensymbol atomformel))
	      (symbol-baum (knoten->praedikatensymbol (binbaum->knoten binbaum))))
	  (cond ((praedikatensymbol=? symbol-neu symbol-baum)
		  (make-binbaum
		   (in-knoten-einfuegen atomformel (binbaum->knoten binbaum))
		   (binbaum->links binbaum)
		   (binbaum->rechts binbaum)))
		((praedikatensymbol<? symbol-neu symbol-baum)
		 (make-binbaum
		  (binbaum->knoten binbaum)
		  (binbaum-einfuegen
		   atomformel
		   (binbaum->links binbaum))
		  (binbaum->rechts binbaum)))
		(else
		 (make-binbaum
		  (binbaum->knoten binbaum)
		  (binbaum->links binbaum)
		  (binbaum-einfuegen
		   atomformel
		   (binbaum->rechts binbaum)))))))))


; Sucht den Baum nach einen Knoten mit einem bestimmten Praedikatensymbol ab.
(define richtiger-knoten
  (lambda (datenbasis atomformel sc fc)
    (binbaum-suche-knoten (datenbasis->datenbaum datenbasis)
			  (atomformel->praedikatensymbol atomformel) sc fc)))

; Macht die eigentliche Arbeit von richtiger-knoten
(define binbaum-suche-knoten
  (lambda (binbaum praedikatensymbol sc fc)
    (if (leerer-binbaum? binbaum)
	fc
	(let ((symbol-knoten 
	       (knoten->praedikatensymbol (binbaum->knoten binbaum))))
	  (cond ((praedikatensymbol=? praedikatensymbol symbol-knoten)
		  (sc (binbaum->knoten binbaum)))
		((praedikatensymbol<? praedikatensymbol symbol-knoten)
		 (binbaum-suche-knoten
		  (binbaum->links binbaum) praedikatensymbol sc fc))
		(else
		 (binbaum-suche-knoten
		  (binbaum->rechts binbaum) praedikatensymbol sc fc)))))))

; Jeder Knoten hat ein Praedikatensymbol, dass aber nicht eigens im Knoten
; sondern in jedem seiner Atomformeln gespeichert ist. Ermittelt das Symbol
; einfach aus der ersten Formel im Knoten.
(define knoten->praedikatensymbol
  (lambda (knoten)
    (atomformel->praedikatensymbol
     (car (knoten->atomliste knoten)))))

; Haengt an einen Knoten eine neue Atomformel an.
(define in-knoten-einfuegen
  (lambda (atomformel knoten)
    (make-knoten (cons 
		  atomformel
		  (knoten->atomliste knoten)))))




