; Enthaelt ein Kopf eine Atomformel?
(define enthaelt-kopf?
  (lambda (kopf)
    (lambda (atomformel)
      (any (lambda (af) (atomformel=? atomformel af))
           (kopf->atomformeln kopf)))))

; Subsumiert ein Kopf einen anderen, d.h. impliziert die Wahrheit eines Kopfes die
; Wahrheit eines anderen?
(define subsumiert?
  (lambda (kopf1)
    (lambda (kopf2)
      (all (enthaelt-kopf? kopf2)
           (kopf->atomformeln kopf1)))))

; Untersucht eine Menge von Koepfen auf gegenseitige Subsumption und
; entfernt Koepfe, die von anderen Koepfen subsumiert werden.
(define subsumierung
  (letrec ((loop (lambda (koepfe)
                   (if (null? koepfe)
                       '()
                       (cons (car koepfe)
                             (loop (remove (subsumiert? (car koepfe)) (cdr koepfe))))))))
    (lambda (koepfe)
      (reverse (loop (reverse (loop koepfe)))))))

; Erzeugt aus einer Datenbasis und einem Kopf einen Strom von Datenbasen,
; die jeweils eine Atomformel des Kopfes enthalten. Dies ersetzt praktisch
; das Backtracking, wie man es vom Basicsatchmo kennt.
(define kopf-einfuegen
  (lambda (kopf)
    (lambda (datenbasis)
      (stream-map (lambda (atomformel) (atomformel-hinzufuegen atomformel datenbasis))
                  (list->stream (kopf->atomformeln kopf))))))

; Fuegt eine ganze Menge von Koepfen ein.
(define koepfe-einfuegen
  (lambda (koepfe datenbasis)
    (foldr (lambda (kopf datenbasen)
             (stream-flatten (stream-map (kopf-einfuegen kopf) datenbasen)))
           (stream datenbasis)
           koepfe)))

; Erzeugt alle Modelle aus einer Spezifikation. Man kann 
; noch als Boolenparameter angeben, ob man die Erzeugung
; am Bildschirm verfolgen will, und ob Optimierung von
; subsumierenden Koepfen erfolgen soll.
(define inferenzmotor
  (lambda (spezifikation trace? subsumieren?)
    (letrec ((loop (lambda (datenbasis)
                     (if trace? (datenbasis-ausgeben "******** Datenbasis *********" datenbasis))
                     (let ((koepfe (ableitbare-koepfe spezifikation datenbasis trace?)))
                       (let ((sauber-koepfe (map kopf-filter koepfe)))
			 (let ((min-koepfe (if subsumieren? 
					       (subsumierung sauber-koepfe)
					       sauber-koepfe)))
			   (if trace? (zeige-subsumption koepfe min-koepfe subsumieren?))
			   (if (null? min-koepfe)
			       (stream datenbasis)
			       (stream-flatten (stream-map loop
							   (koepfe-einfuegen 
							    min-koepfe datenbasis))))))))))
	     loop)))

; Loescht aus einem Kopf alle Builtin-Atomformeln. Diese muessen ohnehin falsch
; sein, da sonst die Regel garnich angewandt waere.
(define kopf-filter
  (lambda (kopf)
    (make-kopf 
     (foldr (lambda (atom liste) (if (builtin? atom) liste (cons atom liste)))
	    '()
	    (kopf->atomformeln kopf)))))
