;--------------------------------------------------------------------------------
; Fehlerbehandlung (hier nur Abbruch)

(define parser-fehler
  (lambda (klassen klasse)
    (error (apply string-append
                  (append (list "Syntax-Fehler: "
                                (symbol->string klasse)
                                " anstelle von ")
                          (unparse-with " oder "
                                        (lambda (klasse accu)
                                          (cons (symbol->string klasse) accu))
                                        klassen
                                        '(".")))))))

(define token-uebrig
  (lambda (token-sequenz)
    (error "Syntax-Fehler: Ueberfluessige Token" token-sequenz)))


;--------------------------------------------------------------------------------
; Ein Parser liefert einen Syntaxbaum und den noch nicht geparsten Rest der Tokensequenz

(define-record parser-antwort (syntax-baum token-sequenz-rest))


;--------------------------------------------------------------------------------
; Der Parser-Puffer ist einfach eine Liste von Token-Infos

(define parser-puffer-erweitern cons)

(define aufrufen-mit-parser-puffer-inhalt
  (lambda (funktion puffer)
    (apply funktion (reverse puffer))))

(define leerer-parser-puffer '())


;--------------------------------------------------------------------------------
; Parser-Aktions-Erzeuger

(define token-puffern
  (lambda (naechste-aktion)
    (lambda (puffer token-sequenz)
      (naechste-aktion (parser-puffer-erweitern (token->info (token-sequenz->kopf token-sequenz))
                                                puffer)
                       (token-sequenz->rest token-sequenz)))))

(define token-ueberlesen
  (lambda (naechste-aktion)
    (lambda (puffer token-sequenz)
      (naechste-aktion puffer
                       (token-sequenz->rest token-sequenz)))))

(define gehe-zu-parser-zustand
  (lambda (zustand)
    (lambda (puffer token-sequenz)
      (let ((naechste-aktion (zustand (token->klasse (token-sequenz->kopf token-sequenz)))))
        (naechste-aktion puffer token-sequenz)))))

(define syntax-baum-bauen
  (lambda (konstruktor)
    (lambda (puffer token-sequenz)
      (make-parser-antwort (aufrufen-mit-parser-puffer-inhalt konstruktor puffer)
                           token-sequenz))))

(define behandle-nt
  (lambda (zustand naechste-aktion)
    (lambda (puffer token-sequenz)
      (let ((antwort ((gehe-zu-parser-zustand zustand) leerer-parser-puffer token-sequenz)))
        (naechste-aktion (parser-puffer-erweitern (parser-antwort->syntax-baum antwort) puffer)
                         (parser-antwort->token-sequenz-rest antwort))))))


;--------------------------------------------------------------------------------

(define parse-spezifikation
  (lambda (klasse)
    (cond ((klasse-in? klasse '(ende))
           (syntax-baum-bauen (konst (make-spezifikation '()))))
          ((klasse-in? klasse '(symbol impliziert))
           (behandle-nt parse-regel
                        (behandle-nt parse-spezifikation
                                     (syntax-baum-bauen (lambda (regel spezifikation)
                                                          (make-spezifikation (cons regel
                                                                                    (spezifikation->regeln spezifikation))))))))
          (else
           (parser-fehler '(ende symbol impliziert) klasse)))))

(define parse-regel
  (lambda (klasse)
    (behandle-nt parse-regelrumpf
                 (behandle-nt parse-regelkopf
                              (syntax-baum-bauen make-regel)))))

(define parse-regelrumpf
  (lambda (klasse)
    (cond ((klasse-in? klasse '(impliziert))
           (token-ueberlesen (syntax-baum-bauen (konst (make-rumpf '())))))
          (else
           (behandle-nt parse-konjunktion
                        (gehe-zu-parser-zustand parse-regelrumpf2))))))

(define parse-regelrumpf2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(impliziert))
           (token-ueberlesen (syntax-baum-bauen make-rumpf)))
          (else
           (parser-fehler '(impliziert) klasse)))))

(define parse-konjunktion
  (lambda (klasse)
    (cond ((klasse-in? klasse '(symbol))
           (behandle-nt parse-atomformel
                        (behandle-nt parse-konjunktion2
                                     (syntax-baum-bauen cons))))
          (else
           (parser-fehler '(symbol) klasse)))))

(define parse-konjunktion2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(und))
           (token-ueberlesen (gehe-zu-parser-zustand parse-konjunktion)))
          (else
           (syntax-baum-bauen list)))))

(define parse-regelkopf
  (lambda (klasse)
    (cond ((klasse-in? klasse '(punkt))
           (token-ueberlesen (syntax-baum-bauen (konst (make-kopf '())))))
          (else
           (behandle-nt parse-disjunktion
                        (gehe-zu-parser-zustand parse-regelkopf2))))))

(define parse-regelkopf2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(punkt))
           (token-ueberlesen (syntax-baum-bauen make-kopf)))
          (else
           (parser-fehler '(punkt) klasse)))))

(define parse-disjunktion
  (lambda (klasse)
    (cond ((klasse-in? klasse '(symbol))
           (behandle-nt parse-atomformel
                        (behandle-nt parse-disjunktion2
                                     (syntax-baum-bauen cons))))
          (else
           (parser-fehler '(symbol) klasse)))))

(define parse-disjunktion2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(oder))
           (token-ueberlesen (gehe-zu-parser-zustand parse-disjunktion)))
          (else
           (syntax-baum-bauen list)))))

(define parse-atomformel
  (lambda (klasse)
    (cond ((klasse-in? klasse '(symbol))
           (token-puffern (behandle-nt parse-termliste
                                       (syntax-baum-bauen (lambda (bezeichner terme)
                                                            (make-atomformel (make-praedikatensymbol bezeichner)
                                                                             terme))))))
          (else
           (parser-fehler '(symbol) klasse)))))

(define parse-termliste
  (lambda (klasse)
    (cond ((klasse-in? klasse '(klammer-auf))
           (token-ueberlesen (behandle-nt parse-terme
                                          (gehe-zu-parser-zustand parse-termliste2))))
          (else
           (syntax-baum-bauen (konst '()))))))

(define parse-termliste2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(klammer-zu))
           (token-ueberlesen (syntax-baum-bauen identitaet)))
          (else
           (parser-fehler '(klammer-zu) klasse)))))

(define parse-terme
  (lambda (klasse)
    (behandle-nt parse-term
                 (gehe-zu-parser-zustand parse-terme2))))

(define parse-terme2
  (lambda (klasse)
    (cond ((klasse-in? klasse '(komma))
           (token-ueberlesen (behandle-nt parse-terme
                                          (syntax-baum-bauen cons))))
          (else
           (syntax-baum-bauen list)))))

(define parse-term
  (lambda (klasse)
    (cond ((klasse-in? klasse '(variablensymbol))
           (token-puffern (syntax-baum-bauen make-variable)))
          ((klasse-in? klasse '(symbol))
           (token-puffern (behandle-nt parse-termliste
                                       (syntax-baum-bauen (lambda (bezeichner terme)
                                                            (make-funktion (make-funktionssymbol bezeichner)
                                                                           terme))))))
          ((klasse-in? klasse '(zahl))
           (token-puffern (syntax-baum-bauen (lambda (zahl)
                                               (make-funktion (make-funktionssymbol zahl) '())))))
          (else
           (parser-fehler '(variablensymbol symbol zahl) klasse)))))


;--------------------------------------------------------------------------------

(define token-sequenz->syntax-baum
  (lambda (token-sequenz)
    (let ((antwort ((gehe-zu-parser-zustand parse-spezifikation) leerer-parser-puffer token-sequenz)))
      (if (equal? (token->klasse (token-sequenz->kopf (parser-antwort->token-sequenz-rest antwort))) 'ende)
          (parser-antwort->syntax-baum antwort)
          (token-uebrig (parser-antwort->token-sequenz-rest antwort))))))
