;--------------------------------------------------------------------------------
; Fehlerbehandlung

(define scanner-fehler
  (lambda (fehler zeichen korrektur)
    (display (string-append "Scanner: " fehler " '" (string zeichen) "'. " korrektur))
    (newline)))


;--------------------------------------------------------------------------------
; Ein Scanner liefert ein Token und den noch nicht gescannten Rest der Zeichensequenz

(define-record scanner-antwort (token zeichen-sequenz-rest))


;--------------------------------------------------------------------------------
; Ein Token besteht aus seiner Klasse und evtl. weiteren Informationen

(define-record token (klasse info))

(define klasse-in? member)

;--------------------------------------------------------------------------------
; Eine Tokensequenz ist ein Strom von Token

(define-syntax make-token-sequenz
  (syntax-rules ()
    ((make-token-sequenz token sequenz)
     (cons-stream token sequenz))))

(define token-sequenz->kopf stream-car)
(define token-sequenz->rest stream-cdr)
(define leere-token-sequenz (stream))
(define zeichen-sequenz-leer? stream-null?)

;--------------------------------------------------------------------------------
; Der Scanner-Puffer ist einfach eine Liste von Zeichen

(define scanner-puffer-erweitern cons)

(define aufrufen-mit-scanner-puffer-inhalt
  (lambda (funktion puffer)
    (funktion (list->string (reverse puffer)))))

(define leerer-scanner-puffer '())


;--------------------------------------------------------------------------------
; Scanner-Aktions-Erzeuger

(define zeichen-puffern
  (lambda (naechste-aktion)
    (lambda (puffer zeichen-sequenz)
      (naechste-aktion (scanner-puffer-erweitern (zeichen-sequenz->kopf zeichen-sequenz)
                                                 puffer)
                       (zeichen-sequenz->rest zeichen-sequenz)))))

(define zeichen-ueberlesen
  (lambda (naechste-aktion)
    (lambda (puffer zeichen-sequenz)
      (naechste-aktion puffer
                       (zeichen-sequenz->rest zeichen-sequenz)))))

(define gehe-zu-scanner-zustand
  (lambda (zustand)
    (lambda (puffer zeichen-sequenz)
      (let ((naechste-aktion (zustand (zeichen-sequenz->kopf zeichen-sequenz))))
        (naechste-aktion puffer zeichen-sequenz)))))

(define token-zurueckgeben
  (lambda (nachbereiter)
    (lambda (puffer zeichen-sequenz)
      (make-scanner-antwort (aufrufen-mit-scanner-puffer-inhalt nachbereiter puffer)
                            zeichen-sequenz))))


;--------------------------------------------------------------------------------

(define scanner-start-zustand
  (lambda (zeichen)
    (cond ((zeichen-whitespace? zeichen)
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-start-zustand)))
          ((char=? zeichen #\%)
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-kommentar-zustand)))
          ((zeichen-kleinbuchstabe? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-symbol-zustand)))
          ((zeichen-grossbuchstabe? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-variable-zustand)))
          ((zeichen-ziffer? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-zahl-zustand)))
          ((char=? zeichen #\+)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-plus-zustand)))
          ((char=? zeichen #\-)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-minus-zustand)))
          ((char=? zeichen #\()
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'klammer-auf))))
          ((char=? zeichen #\))
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'klammer-zu))))
          ((char=? zeichen #\,)
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'komma))))
          ((char=? zeichen #\&)
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'und))))
          ((char=? zeichen #\;)
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'oder))))
          ((char=? zeichen #\.)
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'punkt))))
          ((char=? zeichen ende-zeichen)
           (token-zurueckgeben (interpunktion-nachbereiten 'ende)))
          (else
           (scanner-fehler "Unbekanntes Zeichen" zeichen "Wird ueberlesen.")
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-start-zustand))))))

(define scanner-kommentar-zustand
  (lambda (zeichen)
    (cond ((char=? zeichen #\newline)
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-start-zustand)))
          ((char=? zeichen ende-zeichen)
           (gehe-zu-scanner-zustand scanner-start-zustand))
          (else
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-kommentar-zustand))))))

(define scanner-symbol-zustand
  (lambda (zeichen)
    (cond ((zeichen-bezeichner-rest? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-symbol-zustand)))
          (else
           (token-zurueckgeben symbol-nachbereiten)))))

(define scanner-variable-zustand
  (lambda (zeichen)
    (cond ((zeichen-bezeichner-rest? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-variable-zustand)))
          (else
           (token-zurueckgeben variable-nachbereiten)))))

(define scanner-zahl-zustand
  (lambda (zeichen)
    (cond ((zeichen-ziffer? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-zahl-zustand)))
          (else
           (token-zurueckgeben zahl-nachbereiten)))))

(define scanner-plus-zustand
  (lambda (zeichen)
    (cond ((zeichen-ziffer? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-zahl-zustand)))
          (else
           (scanner-fehler "Erwarte Ziffer anstelle von" zeichen "Fuege '0' ein.")
           (token-zurueckgeben zahl-nachbereiten)))))

(define scanner-minus-zustand
  (lambda (zeichen)
    (cond ((zeichen-ziffer? zeichen)
           (zeichen-puffern (gehe-zu-scanner-zustand scanner-zahl-zustand)))
          ((char=? zeichen #\-)
           (zeichen-ueberlesen (gehe-zu-scanner-zustand scanner-minus-minus-zustand)))
          (else
           (scanner-fehler "Erwarte Ziffer oder '-' anstelle von" zeichen "Fuege '0' ein.")
           (token-zurueckgeben zahl-nachbereiten)))))

(define scanner-minus-minus-zustand
  (lambda (zeichen)
    (cond ((char=? zeichen #\>)
           (zeichen-ueberlesen (token-zurueckgeben (interpunktion-nachbereiten 'impliziert))))
          (else
           (scanner-fehler "Erwarte '>' anstelle von" zeichen "Fuege '>' ein.")
           (token-zurueckgeben (interpunktion-nachbereiten 'impliziert))))))


;--------------------------------------------------------------------------------
; Praedikate fuer die verschiedenen Zeichenklassen

(define zeichen-kleinbuchstabe?
  (lambda (zeichen)
    (member zeichen '(#\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
                      #\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z))))

(define zeichen-grossbuchstabe?
  (lambda (zeichen)
    (member zeichen '(#\A #\B #\C #\D #\E #\F #\G #\H #\I #\J #\K #\L #\M
                      #\N #\O #\P #\Q #\R #\S #\T #\U #\V #\W #\X #\Y #\Z))))

(define zeichen-ziffer?
  (lambda (zeichen)
    (member zeichen '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))))

(define zeichen-spezial?
  (lambda (zeichen)
    (member zeichen '(#\_))))

(define zeichen-bezeichner-rest?
  (lambda (zeichen)
    (or (zeichen-kleinbuchstabe? zeichen)
        (zeichen-grossbuchstabe? zeichen)
        (zeichen-ziffer? zeichen)
        (zeichen-spezial? zeichen))))

(define zeichen-whitespace?
  (lambda (zeichen)
    (member zeichen '(#\space #\newline))))


;--------------------------------------------------------------------------------
; Nachbereiter fuer die verschiedenen Klassen von Token

(define interpunktion-nachbereiten
  (lambda (class)
    (lambda (puffer-inhalt)
      (make-token class 'unspezifiziert))))

(define symbol-nachbereiten
  (lambda (puffer-inhalt)
    (make-token 'symbol puffer-inhalt)))

(define variable-nachbereiten
  (lambda (puffer-inhalt)
    (make-token 'variablensymbol puffer-inhalt)))

(define zahl-nachbereiten
  (lambda (puffer-inhalt)
    (make-token 'zahl (zahl-normalisieren puffer-inhalt))))

(define zahl-normalisieren
  (lambda (str)
    (let ((laenge (string-length str)))
      (if (zero? laenge)
          "0"
          (let ((rest (substring str 1 laenge)))
            (case (string-ref str 0)
              ((#\+) (zahl-normalisieren rest))
              ((#\-) (let ((norm (zahl-normalisieren rest)))
                       (if (string=? norm "0") "0" (string-append "-" norm))))
              ((#\0) (if (> laenge 1) (zahl-normalisieren rest) str))
              (else str)))))))


;--------------------------------------------------------------------------------

(define zeichen-sequenz->token-sequenz
  (letrec ((loop (lambda (start-zustand zeichen-sequenz)
                   (let ((antwort ((gehe-zu-scanner-zustand start-zustand) leerer-scanner-puffer
                                                                           zeichen-sequenz)))
                     (let ((token (scanner-antwort->token antwort)))
                       (make-token-sequenz token
                                           (if (equal? (token->klasse token) 'ende)
                                               leere-token-sequenz
                                               (loop start-zustand
                                                     (scanner-antwort->zeichen-sequenz-rest antwort)))))))))
    (lambda (zeichen-sequenz)
      (loop scanner-start-zustand zeichen-sequenz))))
