; Holt eine Eingabe von der Tastatur.
(define string-einlesen
  (letrec ((loop (lambda (accu)
                   (let ((ch (read-char)))
                     (if (char=? ch #\newline)
                         (list->string (reverse accu))
                         (loop (cons ch accu)))))))
    (lambda ()
      (loop '()))))

; Da die gefundenen Modelle als Stream vorliegen werden einfach
; solange das ersten Element aus dem Stream ausgegeben und ent-
; fernt, bis dieser leer ist.
(define datenbasen-ausgeben
  (lambda (datenbasen)
    (if (stream-null? datenbasen)
        (display "Keine weiteren Modelle.")
        (begin (datenbasis-ausgeben "Folgendes Modell wurde gefunden:" (stream-car datenbasen))
               (display "Weitere Modelle berechnen (j/n)? ")
               (if (member (string-einlesen) '("j" "J" "ja" "Ja" "JA"))
                   (datenbasen-ausgeben (stream-cdr datenbasen))
                   (display-nl "Dann eben nicht..."))))))


; Bekommt einen Zeichenstrom, erzeugt eine Spezifikation, testet auf
; Bereichsbeschraenktheit und startet den Inferenzmotor.
(define modell-erzeugen-aus
  (lambda (eingabe->zeichen-sequenz)
    (lambda (trace? subsumieren? eingabe)
      (let ((spezifikation (token-sequenz->syntax-baum
                            (zeichen-sequenz->token-sequenz
                             (eingabe->zeichen-sequenz eingabe)))))
        (if (bereichsbeschraenkt? spezifikation)
	    (begin 
	      (if trace? (spezifikation-ausgeben spezifikation))
	      (let ((datenbasen ((inferenzmotor spezifikation trace? subsumieren?) leere-datenbasis)))
		(if (stream-null? datenbasen)
		    (display-nl "Die Spezifikation ist widerpruechlich.")
		    (datenbasen-ausgeben datenbasen))))
	    (display-nl "Die Spezifikation ist nicht bereichsbeschraenkt."))))))

; Benutzerschnittstelle fuer Spezifikation in einer Datei
(define modell-erzeugen-aus-datei
  (modell-erzeugen-aus datei->zeichen-sequenz))

; Benutzerschnittstelle fuer Spezifikation in einem String
(define modell-erzeugen-aus-string
  (modell-erzeugen-aus string->zeichen-sequenz))
